@@ -1,3 +1,4 @@
+
 /*
  *	BIRD -- IP Addresses et Cetera for IPv6
  *
@@ -13,6 +14,7 @@
 #include <netinet/in.h>
 #include "lib/string.h"
 #include "lib/bitops.h"
+#include "lib/unaligned.h"
 
 typedef struct ipv6_addr {
   u32 addr[4];
@@ -59,15 +61,22 @@ typedef struct ipv6_addr {
 #define ipa_hton(x) ipv6_hton(&(x))
 #define ipa_ntoh(x) ipv6_ntoh(&(x))
 #define ipa_classify(x) ipv6_classify(&(x))
-/* ipa_opposite and ipa_class_mask don't make sense with IPv6 */
+#define ipa_has_link_scope(x) ipv6_has_link_scope(&(x))
+#define ipa_opposite_m1(x) ({ ip_addr _a=(x); _MI(_I0(_a),_I1(_a),_I2(_a),_I3(_a) ^ 1); })
+#define ipa_opposite_m2(x) ({ ip_addr _a=(x); _MI(_I0(_a),_I1(_a),_I2(_a),_I3(_a) ^ 3); })
+/* ipa_class_mask don't make sense with IPv6 */
 /* ipa_from_u32 and ipa_to_u32 replaced by ipa_build */
 #define ipa_build(a,b,c,d) _MI(a,b,c,d)
 #define ipa_compare(x,y) ipv6_compare(x,y)
 /* ipa_pxlen() requires that x != y */
 #define ipa_pxlen(x, y) ipv6_pxlen(x, y)
 #define ipa_getbit(x, y) ipv6_getbit(x, y)
+#define ipa_put_addr(x, y) ipv6_put_addr(x, y)
 #define ipa_absolutize(x,y) ipv6_absolutize(x,y)
 
+/* In IPv6, SOCK_RAW does not return packet header */
+#define ip_skip_header(x, y) x
+
 ip_addr ipv6_mkmask(unsigned);
 unsigned ipv6_mklen(ip_addr *);
 int ipv6_classify(ip_addr *);
@@ -77,6 +86,11 @@ int ipv6_compare(ip_addr, ip_addr);
 int ipv4_pton_u32(char *, u32 *);
 void ipv6_absolutize(ip_addr *, ip_addr *);
 
+static inline int ipv6_has_link_scope(ip_addr *a)
+{
+  return ((a->addr[0] & 0xffc00000) == 0xfe800000);
+}
+
 /*
  *  This hash function looks well, but once IPv6 enters
  *  mainstream use, we need to check that it has good
@@ -105,6 +119,15 @@ static inline u32 ipv6_pxlen(ip_addr a, ip_addr b)
   return 32 * i + 31 - u32_log2(a.addr[i] ^ b.addr[i]);
 }
 
+static inline byte * ipv6_put_addr(byte *buf, ip_addr a)
+{
+  put_u32(buf+0,  _I0(a));
+  put_u32(buf+4,  _I1(a));
+  put_u32(buf+8,  _I2(a));
+  put_u32(buf+12, _I3(a));
+  return buf+16;
+}
+
 /*
  *  RFC 1883 defines packet precendece, but RFC 2460 replaces it
  *  by generic Traffic Class ID with no defined semantics. Better
